/**
 * @file SO1602AWWB.cpp
 * @author たま吉さん
 * @brief 有機ＥＬキャラクタディスプレイモジュール 制御関数
 */


#include "SO1602AWWB.h"
#include "Wire.h"

#define DEVADR 0x3c  // デバイススレーブアドレス

/**
 * @brief OLED キャラクタディスプレイの初期化
 * @return uint8_t 送信結果　
 */
uint8_t OLEDinit() {
  Wire.beginTransmission(DEVADR);
  Wire.write(0x00); // データ送信
  Wire.write(0x01); // Clear Display (表示部をクリア）
  Wire.write(0x02); // Return Home (カーソル位置を先頭に移動）
  Wire.write(0x0c); // Send Display on command (表示ON、カーソル表示なし)
  Wire.write(0x01); // Clear Display (表示部をクリア、取りあえず説明書通りにもう一度クリア）  
  return Wire.endTransmission();
}

/**
 * @brief OLED キャラクタディスプレイ コントラスト設定
 * @param cont コントラスト値
 * @return uint8_t 送信結果
 */
uint8_t OLEDcontrast(int cont){
  Wire.beginTransmission(DEVADR);
  Wire.write(0x00); // データ送信
  Wire.write(0x2a); // RE=1
  Wire.write(0x79); // SD=1
  Wire.write(0x81); // コントラストセット
  Wire.write(cont); // コントラスト値
  Wire.write(0x78); // SDを0に戻す
  Wire.write(0x28); // 2C=高文字　28=ノーマル
  return Wire.endTransmission();   
}

/**
 * @brief OLED キャラクタディスプレイ カーソル移動
 * @param x X軸位置
 * @param y Y軸位置
 * @return uint8_t 送信結果
 */
uint8_t OLEDPos(uint8_t x, uint8_t y) {
  uint8_t adr = 0x80+y*32+x;
  Wire.beginTransmission(DEVADR);
  Wire.write(0x00); // データ送信
  Wire.write(adr);  // アドレス
  return Wire.endTransmission();
}

/**
 * @brief OLED キャラクタディスプレイ 文字列の表示
 * @param text 表示文字列
 * @return uint8_t 送信結果
 */
uint8_t OLEDprint(uint8_t* text) {
  Wire.beginTransmission(DEVADR);
  Wire.write(0x40);                      // Set DDRAM Address
  Wire.write(text,strlen((char*)text));  // 表示データ
  return Wire.endTransmission();
}

/**
 * @brief OLED キャラクタディスプレイ 文字の表示
 * @param c 表示文字
 * @return uint8_t 送信結果
 */
uint8_t OLEDputch(uint8_t c) {
  Wire.beginTransmission(DEVADR);
  Wire.write(0x40);               // Set DDRAM Address
  Wire.write(c);                  // 表示データ
  return Wire.endTransmission();
}

/**
 * @brief OLED キャラクタディスプレイ 表示のクリア
 * @return uint8_t 送信結果
 */
uint8_t OLEDcls() {
  Wire.beginTransmission(DEVADR);
  Wire.write(0x00); // データ送信
  Wire.write(0x01); // Clear Display (表示部をクリア）
  Wire.write(0x02); // Return Home (カーソル位置を先頭に移動）  
  Wire.write(0x01); // Clear Display (表示部をクリア）
  return Wire.endTransmission();   
}

/**
 * @brief OLED キャラクタディスプレイ 表示のON/OFF指定
 * @param sw 表示のON/OFF指定
 * @return uint8_t 送信結果
 */
uint8_t OLEDisplay(uint8_t sw) {
  uint8_t mode = 0b1000 | ((sw?1:0)<<2);
  Wire.beginTransmission(DEVADR);
  Wire.write(0x00); // データ送信
  Wire.write(mode); // Display ON/OFF
  return Wire.endTransmission();     
}

/**
 * @brief OLED キャラクタディスプレイ カーソル表示指定
 * @param sw カーソルの表示のON/OFF指定
 * @return uint8_t 送信結果
 */
uint8_t OLEDCurs(uint8_t sw) {
  uint8_t mode = 0b1100 | (sw?3:0);
  Wire.beginTransmission(DEVADR);
  Wire.write(0x00); // データ送信
  Wire.write(mode); // Display ON/OFF
  return Wire.endTransmission();       
}
