/**
 * @file charinout.cpp
 * @brief character in/out to device
 */
#include <stdint.h>
#include "device.h"
#include "mcurses.h"
#include "console.h"
#include "SO1602AWWB.h"
#include "SC1602BS.h"
#include "serial.h"
#include "ttconfig.h"
#include "charinout.h"

//***************
//* 文字入出力 ***
//***************
/**
 * @brief ターミナルにBELLコードを出力してBEEP音を鳴らす
 */
void c_beep() {
  addch(0x07);
}

#if USE_TEST_OP == 1
/**
 * @brief ターミナル表示の改行
 */
void c_newLine() {
  
  Arduino_putchar('\n');
}
/**
 * @brief ターミナルに1文字出力 CURSESをPASSして直接出力
 * @details Basicの機能確認モード　CURSUSをパスする
 * @param c 　ターミナルに出力する文字
 */
void c_addch(uint8_t c) {
    Arduino_putchar(c);
}

/**
 * @brief ターミナルから1文字入力
 * @return uint8_t 　ターミナルから CURSUSをPASSして直接取得
 * @details Basicの機能確認モード　CURSUSをパスする
 */
uint8_t c_getch() {
  return Arduino_getchar();
}

#else
/**
 * @brief ターミナル表示の改行
 */
void c_newLine() {
  int16_t x,y;
  getyx(y,x);
  if (y>=MCURSES_LINES-1) {
    scroll();
    move(y,0);
  } else {
    move(y+1,0);  
  }  
}

/**
 * @brief ターミナルに1文字出力
 * @param c 　ターミナルに出力する文字
 */
void c_addch(uint8_t c) {
  if (c=='\n')
    c_newLine();
  else
    addch(c);
}

/**
 * @brief ターミナルから1文字入力
 * @return uint8_t 　ターミナルから取得した文字
 */
uint8_t c_getch() {
  return getch();
}
#endif

/**
 * @brief 指定デバイスへの1文字出力
 * @param c  出力する文字
 * @param devno 出力デバイス番号
 * @todo OP_TEST==1 の時に devno==CDEV_SCREEN で m_cursusをパスする処理の追加
 */
void c_putch(uint8_t c, uint8_t devno) {
  if (devno == CDEV_SCREEN ) {
     c_addch(c);       // 標準出力 
  } else if (devno == CDEV_MEMORY) {
     mem_putch(c);     // メモリー出力
  } 
#if USE_SO1602AWWB == 1
  else if (devno == CDEV_CLCD) {
    OLEDputch(c);      // OLEDキャラクタディスプレイ出力 I2C
  }
#endif
#if USE_SC1602BS == 1
  else if (devno == CDEV_LCD_CH) {
      LCDputch(c);     // LCDキャラクタディスプレイ出力 4bit
  }

#endif
} 

/**
 * @brief 文字列出力
 * @param s 出力文字列
 * @param devno 出力デバイス番号
 */
void c_puts_devno(const char *s, uint8_t devno) {
  while (*s) c_putch(*s++, devno); //終端でなければ出力して繰り返す
}

/**
 * @brief CDEV_SCREEN に文字列出力 
 * @param s 出力文字列
 */
void c_puts(const char *s)
{
  while (*s) c_putch(*s++, CDEV_SCREEN); //終端でなければ出力して繰り返す
}
