/**
 * @file console.cpp
 * @brief コンソール関係（スクリーン制御、ラインエディタ等）
 * @author 2019/06/08 たま吉さん
 * @note 全角処理は全く修正していないし、動作確認もしてない
 * @note basic.cpp の関数から移動したものがある
 * @note LCDの処理も追加してある。
 */

#include "Arduino.h"
#include "basic.h"
#include "charlib.h"
#include "charinout.h"
#include "SC1602BS.h"
#include "intercode.h"
#include "errcode.h"
#include "sub.h"
#include "gpio.h"
#include "ctrlcode.h"
#include "table.h"
#include "keyword.h"
#include "getparam.h"
#include "device.h"
#include "serial.h"
#include "ttconfig.h"
#include "console.h"

static int16_t lbuf_pos = 0;        // コマンドライン内参照位置
uint8_t lbuf[SIZE_LINE];            // コマンドラインバッファ

// mcursesライブラリ(修正版)
// オリジナル版を修正して組み込み：https://github.com/ChrisMicro/mcurses
#include "mcurses.h"

// ** basic.cpp から移動　begin
/**
 * @fn char sstyle(uint8_t code,const uint8_t *table, uint8_t count) 
 * @brief LIST出力整形例外チェック関数
 * @param  code チェック対象の関数
 * @param  table 中間コードテーブル
 * @param  count 中間コードの数
 * @note 間接的に console.cpp から呼ばれている nospacea()
 */
char sstyle(uint8_t code,const uint8_t *table, uint8_t count) {
  while(count--) //中間コードの数だけ繰り返す
    if (code == pgm_read_byte(table + count)) //もし該当の中間コードがあったら
      return 1; //1を持ち帰る
  return 0; //（なければ）0を持ち帰る
}

//*** LIST出力整形例外チェック マクロ関数 ************

static char nospacea(uint8_t c) { return sstyle(c, i_nsa, sizeof(i_nsa)); }
static char nospaceb(uint8_t c) { return sstyle(c, i_nsb, sizeof(i_nsb)); }
// 必ず前に空白を入れる中間コードか？
static char spacef(uint8_t c)   { sstyle(c, i_sf, sizeof(i_sf)); }
// 後ろが変数、数値、定数の場合、後ろに空白を空ける中間コードか
static char spacebifValue(uint8_t c) { sstyle(c, i_sb_if_value, sizeof(i_sb_if_value));}

// ** basic.cpp から移動 end *************

uint8_t flgCurs = 0;


/**
 * @brief メモリへの文字出力
 * @param c  出力する文字
 * @note ※コマンドラインバッファをワーク用に利用
 */
void mem_putch(uint8_t c) {
  if (lbuf_pos < SIZE_LINE) {
   lbuf[lbuf_pos] = c;
   lbuf_pos++;
  }
}

/**
 * @brief メモリ書き込みポインタのクリア
 * @note  ※コマンドラインバッファをワーク用に利用
 */
void clearlbuf() {
  lbuf_pos=0;
  memset(lbuf,0,SIZE_LINE);
}

/**
 * @brief 改行
 * @param devno デバイス番号
 * @todo CDEV_SCREEN && OP_TEST == 1 で mcursus を使わない処理を追加
 */
void newline_dev(uint8_t devno) {
  if ( devno == CDEV_SCREEN ) {
    c_newLine();     // 標準出力 
  } else if (devno == CDEV_MEMORY ) {
    mem_putch('\n'); // メモリー出力
  }
  else if (devno == CDEV_LCD_CH) {
    LCDnewline();
  }
}

/**
 * @brief 改行
 * @todo OP_TEST == 1 で mcursus を使わない処理を追加
 */
void newline(void) {
    c_newLine();     // 標準出力 
}

/**
 * @brief コンソール初期化
 */
void init_console() {
  // mcursesの設定
  setFunction_putchar(Arduino_putchar);  // 依存関数
  setFunction_getchar(Arduino_getchar);  // 依存関数
  // Basic test modeでは cursusをパスする
  #if USE_TEST_OP == 0
  initscr();                             // 依存関数
  setscrreg(0,MCURSES_LINES);  
  #endif
}
 
/**
 * @brief カーソル表示制御
 * @param mode 
 * @note curs_set() import from mcurses.c
 */
void c_show_curs(uint8_t mode) {
  flgCurs = mode;
#if USE_TEST_OP == 0
  curs_set(mode);
#endif
}

/**
 * @brief 画面クリア
 */
void icls() {
#if USE_TEST_OP == 0
  clear();
  move(0,0);
#endif
}

/**
 * @brief カーソル移動 LOCATE x,y
 */
void ilocate() {
  int16_t x,  y;

  if ( getParamInt16(&x, true) || getParamInt16(&y, false)) return;
  if ( x >= c_getWidth() )   // xの有効範囲チェック
     x = c_getWidth() - 1;
  else if (x < 0)  x = 0;  
  if( y >= c_getHeight() )   // yの有効範囲チェック
     y = c_getHeight() - 1;
  else if(y < 0)   y = 0;

  // カーソル移動
  move(y,x);
}

// 文字色/属性テーブル
const PROGMEM  uint16_t tbl_fcolor[]  =
   { F_BLACK,F_RED,F_GREEN,F_BROWN,F_BLUE,F_MAGENTA,F_CYAN,A_NORMAL,F_YELLOW};
const PROGMEM  uint8_t tbl_attr[]  =
    { A_NORMAL, A_UNDERLINE, A_REVERSE, A_BLINK, A_BOLD };  

/**
 * @brief 文字色の指定 COLOLR fc[,bc]
 */
void icolor() {
  int16_t fc,  bc = 0;
  if ( getParamRange(&fc, 0, 8, false) ) return;
  if(*cip == I_COMMA) {
    // 第2引数(背景色）の指定あり
    cip++;
    if ( getParamRange(&bc, 0, 8, false) ) return;  
  }
  attrset(pgm_read_word(&tbl_fcolor[fc])|(pgm_read_word(&tbl_fcolor[bc])<<4));
}

/**
 * @brief 文字属性の指定 ATTRコマンド
 */
void iattr() {
  int16_t attr;
  if ( getParamRange(&attr, 0, 4, false) ) return;
  attrset(pgm_read_byte(&tbl_attr[attr]));
}

/**
 * @brief 文末空白文字のトリム処理
 * @param str トリム対象の文字列
 * @return char*  パラメータのstrのポインタを返す
 */
char* tlimR(char* str) {
  uint16_t len = strlen(str);
  for (uint16_t i = len - 1; i>0 ; i--) {
    if (str[i] != ' ') break;
    str[i] = 0;
  }
  return str;
}

// http://katsura-kotonoha.sakura.ne.jp/prog/c/tip00010.shtml
//*********************************************************
// 文字列 str の str[nPos] について、
//   ０ …… １バイト文字
//   １ …… ２バイト文字の一部（第１バイト）
//   ２ …… ２バイト文字の一部（第２バイト）
// のいずれかを返す。
//*********************************************************
#define jms1(c) (((0x81<=c)&&(c<=0x9F))||((0xE0<=c)&&(c<=0xFC))) 
#define jms2(c) ((0x7F!=c)&&(0x40<=c)&&(c<=0xFC))
uint8_t isJMS( uint8_t *str, uint16_t nPos ) {
  uint16_t i;
  uint8_t state; // { 0, 1, 2 }

  state = 0;
  for( i = 0; str[i] != '\0'; i++ ) {
    if      ( ( state == 0 ) && ( jms1( str[i] ) ) ) state = 1; // 0 -> 1
    else if ( ( state == 1 ) && ( jms2( str[i] ) ) ) state = 2; // 1 -> 2
    else if ( ( state == 2 ) && ( jms1( str[i] ) ) ) state = 1; // 2 -> 1
    else                                             state = 0; // 2 -> 0, その他
    // str[nPos] での状態を返す。
    if ( i == nPos ) return state;
  }
  return 0;
}
 
/**
 * @brief バッファ内の指定位置の文字を削除する
 * @param str バッファのポインタ
 * @param pos 削除する文字の位置
 */
void line_delCharAt(uint8_t* str, uint8_t pos) {
  uint8_t len = strlen((char *)str); // 文字列長さ 
  if (pos == len-1) {
    // 行末の場合
     str[pos] = 0;
   } else {
    memmove(&str[pos], &str[pos+1],len-1-pos); 
    str[len-1] = 0;
  }
}
 
/**
 * @brief 1行バッファ内の指定位置に文字を挿入する
 * @param str バッファの文字列ポインタ
 * @param pos 文字を挿入する指定位置
 * @param c 　挿入文字
 */
void line_insCharAt(uint8_t* str, uint8_t pos,uint8_t c) {
  uint8_t len = strlen((char *)str); // 文字列長さ 
  if (len >= SIZE_LINE-1)
    return;
    
  if (pos == len) {
    // 行末の場合
     str[pos] = c;
     str[pos+1] = 0;
   } else {
    memmove(&str[pos+1], &str[pos],len-pos);
    str[pos] = c;
    str[len+1] = 0;
  }
}

/**
 * @brief カーソルを１文字前に移動
 * @param str 該当行の文字列ポインタ
 * @param ln  該当行の長さ
 * @param pos 移動前のカーソルの位置
 * @return pos 移動後のカーソルの位置
 */
uint8_t line_movePrevChar(uint8_t* str, uint8_t ln, uint8_t pos) {     
  if (pos > 0) {
    pos--;
  }
  move(ln,pos+1);
  return pos;
}

/**
 * @brief カーソルを１文字次に移動
 * @param str 該当行の文字列ポインタ
 * @param ln  該当行の長さ
 * @param pos 移動前のカーソルの位置
 * @return pos 移動後のカーソルの位置
 */
uint8_t line_moveNextChar(uint8_t* str, uint8_t ln, uint8_t pos) {
  uint8_t len = strlen((char *)str); // 文字列長さ 
  if (pos < len ) {
    pos++;
  }
  move(ln,pos+1);  
  return pos;
}

/**
 * @brief ラインエディタ内容の再表示 
 * @param ln  行番号
 * @param pos 位置
 * @param len 行の長さ
 */
void line_redrawLine(uint8_t ln,uint8_t pos,uint8_t len) {
  move(ln,0);
  deleteln();
  c_putch('>', CDEV_SCREEN);
  c_puts((char *)lbuf);
  move(ln,pos+1);  
}

/**
 * @brief 中間コード格納行の行番号取得(1行分リスト内行番号取得)
 * @param *lp  中間コードバッファのポインタ
 * @return 戻り値 行番号
 * @note export basic.cpp console.cpp
 */
int16_t getlineno(uint8_t *lp) {
  return  (*lp == 0) ? -1: *(lp + 1) | *(lp + 2) << 8; //行番号を持ち帰る
}

/**
 * @brief 指定行番号のリストポインタを取得
 * @param lineno  指定行番号
 * @return 戻り値 指定行番号のリストポインタ
 * @note listbuf  extern で宣言済み
 */
uint8_t* getlp(short lineno) {
  uint8_t *lp; // ポインタ
  for (lp = listbuf; *lp && getlineno(lp) < lineno; lp += *lp); // 先頭から末尾まで繰り返す
  return lp; // ポインタを持ち帰る
}

/**
 * @brief 指定した行のプログラムテキストを取得する
 * @param lineno  指定行番号
 * @return 戻り値 指定行番号のリストポインタ
 */
char* getLineStr(int16_t lineno) {
    uint8_t* lp = getlp(lineno);
    if (lineno != getlineno(lp)) 
      return NULL;
    
    // 行バッファへの指定行テキストの出力
    clearlbuf(); // バッファのクリア
    putnum(lineno, 0,3); // 行番号を表示
    c_putch(' ',3);      // 空白を入れる
    putlist(lp+3,3);     // 行番号より後ろを文字列に変換して表示
    c_putch(0,3);        // \0を入れる
    return (char *)lbuf;
}

/**
 * @brief 指定した行の前の行番号を取得する
 * @param lineno  指定行番号
 * @return 戻り値 指定行番号の前の行番号
 * @note listbuf  extern で宣言済み
 */
int16_t getPrevLineNo(int16_t lineno) {
  uint8_t* lp, *prv_lp = NULL;
  int16_t rc = -1;
  for ( lp = listbuf; *lp && (getlineno(lp) < lineno); lp += *lp) {
    prv_lp = lp;
  }
  if (prv_lp)
    rc = getlineno(prv_lp);
  return rc;
}

/**
 * @brief 指定した行の次の行番号を取得する
 * @param lineno 指定行番号
 * @return 戻り値 指定行番号の次の行番号
 */
int16_t getNextLineNo(int16_t lineno) {
  uint8_t* lp;
  int16_t rc = -1;
  
  lp = getlp(lineno); 
  if (lineno == getlineno(lp)) { 
    // 次の行に移動
    lp+=*lp;
    rc = getlineno(lp);
  }
  return rc;
}

/**
 * @brief ラインエディタ
 * @note 全角文字の処理は対応も確認もしていない。
 */
void c_gets() {
  uint8_t x,y;
  uint8_t c;               // 文字
  uint8_t len = 0;         // 文字数
  uint8_t pos = 0;         // 文字位置
  int16_t line_index = -1; // リスト参照
  int16_t tempindex;
  uint8_t *text;
  int16_t value, tmp;

  getyx(y,x);
  clearlbuf();
  while ((c = c_getch()) != KEY_CR) { //改行でなければ繰り返す
    if (c == KEY_TAB) {  // [TAB]キー
      if  (len == 0) {
        if (errorLine > 0) {
          // 空白の場合は、直前のエラー発生行の内容を表示する
          text = (uint8_t*)tlimR((char*)getLineStr(errorLine));
        } else {
          text = (uint8_t*)tlimR((char*)lbuf);          
        }
        if (text) {
          // 指定した行が存在する場合は、その内容を表示する
          //strcpy(lbuf,text);
          pos = 0;
          len = strlen((char*)lbuf);
          line_redrawLine(y,pos,len);
        } else {
          c_beep();
        }   
      } else if (len) {
        // 先頭に数値を入力している場合は、行の内容を表示する
        text = lbuf;
        value = 0; //定数をクリア
        tmp   = 0; //変換過程の定数をクリア
        if (c_isdigit(*text)) { //もし文字が数字なら
           do { //次の処理をやってみる
              tmp = 10 * value + *text++ - '0'; //数字を値に変換
              if (value > tmp) { // もし前回の値より小さければ
                value = -1;
                break;
               }
               value = tmp;
            } while (c_isdigit(*text)); //文字が数字である限り繰り返す
         }
         if (value >0) {
            text = (uint8_t*)tlimR(getLineStr(value));
            if (text) {
              // 指定した行が存在する場合は、その内容を表示する
              //strcpy(lbuf,text);
              pos = 0;
              len = strlen((char*)lbuf);
              line_redrawLine(y,pos,len);
              line_index = value;
            } else {
              c_beep();
            }
         }
      }
    } else if ( (c == KEY_BACKSPACE) && (len > 0) && (pos > 0) ) { // [BS]キー
      //if (pos > 1 && isZenkaku(lbuf[pos-2])) { // 2019/11/14 全角判定ミス対応
      if ( (pos > 1 ) && (isJMS(lbuf,pos-1) == 2) ) {

        // 全角文字の場合、2文字削除
        pos-=2;
        len-=2;
        line_delCharAt(lbuf, pos);
        line_delCharAt(lbuf, pos);
      } else {      
        len--; pos--;
        line_delCharAt(lbuf, pos);
      }
      line_redrawLine(y,pos,len);
    } else if ( (c == KEY_DC) && (len > 0) && (pos < len) ) {     // [Delete]キー
      if (isZenkaku(lbuf[pos])) {
        // 全角文字の場合、2文字削除
        line_delCharAt(lbuf, pos);
        len--;
      }
      line_delCharAt(lbuf, pos);
      len--;
      line_redrawLine(y,pos,len);      
    } else if (c == KEY_LEFT && pos > 0) {                // ［←］: カーソルを前の文字に移動
      pos = line_movePrevChar(lbuf,y,pos);
    } else if (c == KEY_RIGHT && pos < len) {             // ［→］: カーソルを次の文字に移動
      pos = line_moveNextChar(lbuf,y,pos);
    } else if (  (c == KEY_PPAGE || c == KEY_UP )   || // ［PageUP］: 前の行のリストを表示
                 (c == KEY_NPAGE || c == KEY_DOWN ) ){ // ［PageDown］: 次の行のリストを表示 
      if (line_index == -1) {
        line_index = getlineno(listbuf);
        tempindex = line_index;
      } else {
        if (c == KEY_PPAGE || c == KEY_UP)
          tempindex = getPrevLineNo(line_index);
        else 
          tempindex = getNextLineNo(line_index);
      }
      if (tempindex > 0) {
        line_index = tempindex;
        text = (uint8_t*)tlimR(getLineStr(line_index));
        //strcpy((char *)lbuf, (char *)text);
        pos = 0;
        len = strlen((char *)lbuf);
        line_redrawLine(y,pos,len);
      } else {
        pos = 0; len = 0;
        memset(lbuf,0,SIZE_LINE);
        line_redrawLine(y,pos,len);
      }
    } else if (c == KEY_HOME ) { // [HOME]キー
       pos = 0;
       move(y,1);
    } else if (c == KEY_END ) {  // [END]キー
       pos = len;
       move(y,pos+1);
    } else if (c == KEY_F1 ) {   // [F1]キー(画面クリア)
       icls();
       pos = 0; len = 0;
       getyx(y,x);
       memset(lbuf,0,SIZE_LINE);
       c_putch('>',CDEV_SCREEN);
    } else if (c == KEY_F2 ) {   // [F2]キー(ラインクリア)
       pos = 0; len = 0;
       memset(lbuf,0,SIZE_LINE);
       line_redrawLine(y,pos,len);
    } else if (c>=32 && (len < (SIZE_LINE - 1))) { 
      //表示可能な文字が入力された場合の処理（バッファのサイズを超えないこと）      
      line_insCharAt(lbuf,pos,c);
      insch(c);
      pos++;len++;
    }
  }
  newline(); //改行
  lbuf[len] = 0; //終端を置く

  if (len > 0) { //もしバッファが空でなければ
    while (c_isspace(lbuf[--len])); //末尾の空白を戻る
    lbuf[++len] = 0; //終端を置く
  }
}

/**
 * @brief 実行時の強制的な中断の判定
 * @details ESCかCTRL-Cキーで実行中断の判定
 * @return err  break: ERR_CTR_C  nobreak: 設定しない
 * @todo エラーコードではなく、 true/falseを返すべき
 */
uint8_t isBreak() {
  uint8_t c = c_kbhit();
  if (c) {
      // [ESC],［CTRL_C］キーなら中断
      if (c == CHAR_CTRL_C || c == CHAR_ESCAPE ) { 
        err = ERR_CTR_C;  // エラー番号をセット
        prevPressKey = 0;
      } else {
        prevPressKey = c;
      }
   }
   return err;
}


/**
 * @fn putlist(uint8_t* ip, uint8_t devno)
 * @brief 指定位置プログラムリスト出力
 * @details 中間コードをキーワード変換し、１行分を出力する
 * @param ip 中間コート格納ポインタ
 * @param devno デバイス番号
 * @note export basic.cpp console.cpp files.cpp import console.cpp as spacebifValue 
 *       sub.cpp as putnum putHexnum putBinnum
 */
void putlist(uint8_t* ip, uint8_t devno ) {
  uint8_t i; //ループカウンタ

  // 行末まで繰り返し処理
  while (*ip != I_EOL) { 
  
    // 中間コード・キーワードの処理
    if (*ip < SIZE_KWTBL) {
      // もしキーワードなら、キーワードテーブルの文字列を出力
      c_puts_devno((char*)pgm_read_word(&(kwtbl[*ip])),devno);
      // 次の中間コードが':'でない場合、ルールをチェックし、キーワードの後ろに空白を出力
      if (*(ip+1) != I_COLON) 
        if ( ((!nospacea(*ip) || spacef(*(ip+1))) && (*ip != I_COLON) && (*ip != I_SQUOT))
        || ((*ip == I_CLOSE)&& (*(ip+1) != I_COLON  && *(ip+1) != I_EOL && !nospaceb(*(ip+1)))) 
        || ( spacebifValue(*ip) && spacebifValue(*(ip+1))) ) // もし例外にあたらなければ
           c_putch(' ',devno); // 空白を出力

      // キーワードがコメント("REM" or "'")の場合、以降の文字列をそのまま出力する
      if (*ip == I_REM||*ip == I_SQUOT) {
        ip++;                   // ポインタを文字数へ進める
        i = *ip++;              // 文字数を取得してポインタをコメントへ進める
        while (i--)             // 文字数だけ繰り返す
          c_putch(*ip++,devno); // ポインタを進めながら文字を出力
        return;
      }
      ip++;

    // 数値定数の処理
    } else if (*ip == I_NUM) {      
      // 数値を出力
      ip++;  putnum(*ip | *(ip + 1) << 8, 0,devno); 

      ip += 2;               // ポインタを次の中間コードへ進める
      if (!nospaceb(*ip))    // もし例外にあたらなければ
        c_putch(' ',devno);  // 空白を出力

    // 16進数定数の処理
    } else if (*ip == I_HEXNUM) {
      // 16進数定数を出力($+16進数定数)
      ip++; c_putch('$',devno); putHexnum(*ip | *(ip + 1) << 8, 2,devno);

      ip += 2;              // ポインタを次の中間コードへ進める
      if (!nospaceb(*ip))   // もし例外にあたらなければ
        c_putch(' ',devno); // 空白を出力

    // 2進定数の処理（値に応じて8桁 or 16桁出力）
    } else if (*ip == I_BINNUM) {
      ip++; c_putch('`',devno); //"`"を出力
      if (*(ip + 1))            // 2バイト目が1以上
          putBinnum(*ip | *(ip + 1) << 8, 16,devno); // 16桁で出力
      else
          putBinnum(*ip , 8,devno);                  // 8桁で出力

      ip += 2;                 // ポインタを次の中間コードへ進める
      if (!nospaceb(*ip))      // もし例外にあたらなければ
        c_putch(' ',devno);    // 空白を出力

    // 変数の処理  
    } else if (*ip == I_VAR) {
      ip++; //ポインタを変数番号へ進める
      c_putch(*ip++ + 'A',devno); // 変数名を取得して出力
      if (!nospaceb(*ip))         // もし例外にあたらなければ
        c_putch(' ',devno);       // 空白を出力

    // 文字列の処理    
    } else if (*ip == I_STR) {
      char c; // 文字列の括りに使われている文字（「"」または「'」）
      // 文字列の括りに使われている文字を調べる
      c = '\"';                  // 文字列の括りを仮に「"」とする
      ip++;                      // ポインタを文字数へ進める
      for (i = *ip; i; i--)      // 文字数だけ繰り返す
        if (*(ip + i) == '\"') { // もし「"」があれば
          c = '\'';              // 文字列の括りは「'」
          break;                 // 繰り返しを打ち切る
        }

      // 文字列を出力する
      c_putch(c,devno);          // 文字列の括りを出力
      i = *ip++;                 // 文字数を取得してポインタを文字列へ進める
      while (i--)                // 文字数だけ繰り返す
        c_putch(*ip++,devno);    // ポインタを進めながら文字を表示
      c_putch(c,devno);          // 文字列の括りを表示

      // もし次の中間コードが変数、ELSEだったら空白を出力
      if (*ip == I_VAR || *ip ==I_ELSE) 
        c_putch(' ',devno);

    //どれにも当てはまらなかった場合
    } else { 
      err = ERR_SYS; // エラー番号をセット
      return;        // 終了する
    }
  }
}

