/**
 * @file flash.cpp
 * @author your name (you@domain.com)
 * @brief 内部フラッシュEEPROMの制御ルーチン
 * @version 0.1
 * @date 2022-02-10
 * 
 * @copyright Copyright (c) 2022
 * 
 */
#include <Arduino.h>
#include "ttconfig.h"
#include "workmemdef.h"

#include "flash.h"
#if USE_HS_EEPROM == 1

#include "utility/data_flash_util.h"

// ページとは ロードセーブするときの番号をページ番号とここでは定義している。単にload/save だとページ0になる
typedef enum {
	PFDS_GET,			// 保存ページ番号を取得 有効なページかの判定を行い、無効ならエラー表示に移行
    PFDS_USABLE,        // 保存ページの対象ブロックが使用可能か判定（現在は常に可） 
    PFDS_SET,           // 開始ブロック、ブロック数を設定する
    PFDS_WADDR,         // 内部データフラッシュ書き込み開始アドレス設定
	PFDS_WOPEN,	        // データフラッシュ処理のオープン
	PFDS_WBLANK,	    // ブランクかチェックする。ブランクならPFDS_WRITE、違えば PFDS_ERASE
	PFDS_WERASE,		// 書き込み対象のブロックをイレーズする
	PFDS_WRITE,			// 対象ブロックに書き込みを行う	
    PFDS_WCONT,         // 最終ブロック: PFDS_COMPL 途中ブロック：ブロック数+1 PFDS_WRITE
	PFDS_WCLOSE,	    // 終了処理
    PFDS_RADDR,
    PFDS_ROPEN,         // Read Open
    PFDS_READ,          // Read one page
    PFDS_RCONT,         // increment block counter
    PFDS_RCLOSE,        // 終了処理
	PFDS_ERR,           // エラー内容をシリアル出力
} sdlSeq ;
// unit16_t romAddr  dataflash内のアドレス 0-8092 
/**
 * @fn void eeprom_update_pages(uint8_t* buffa, uint8_t pageNo, size_t size)
 * @brief FlashDataRomの指定したページの読み出しを行う
 * @param [in] wrBuffa   FlashDataRom から読み出したデータを保存するbuffa
 * @param [in] pageNo  page numberを指定する
 * @note ソースのサイズに関係なく、1ページ全部にデータを書き込んでいる。
 * @attention Multi Block (ソースに1Kbyte以外 2K, 4K, 8K) 指定の時だけ有効な関数
 */
void eeprom_update_pages(uint8_t* wrBuffa, uint8_t pageNo)
{
    sdlSeq seq = PFDS_USABLE;           //! シーケンス変数

    bool endFlag = false;               //! シーケンス終了フラグ
    int  blockStart = 0;                //! 書き込み開始ブロックの番号
    int  blockCounter = 0;              //! Block Conter　書き込み中のブロックの番号
    int  blockEnd = 0;                  //! 書き込み終了ブロックの番号
    uint16_t flashAddress = 0;          //! 内部データフラッシュ書き込み開始アドレス 
    int result=0;

    do {
        switch (seq)
        {
        case PFDS_USABLE:       // 将来ここで使用可能なブロックかどうかの判定を行う
            seq = PFDS_SET;
            break;

        case PFDS_SET:          
            blockStart = pageNo * EEPROM_PER_BLOCK;    // 開始ブロックを設定
            blockEnd = blockStart + EEPROM_PER_BLOCK;  // 終了
            blockCounter = blockStart;
            seq = PFDS_WOPEN;
            break;

        case PFDS_WOPEN:
            result = pfdl_open();
            if (result == PFDL_OK) {
                seq = PFDS_WADDR;
            } 
            else {
                Serial.print("DataFlash Read error in PFDS_WOPEN : ");
                Serial.print(result);                
                seq = PFDS_ERR;
            }
            break;

        case PFDS_WADDR:
            flashAddress = blockStart * PFDL_DATA_FLASH_BLOCK_SIZE;
            seq = PFDS_WBLANK;
            break;

        case PFDS_WBLANK:
            result = pfdl_blank_check( (uint16_t)flashAddress, PFDL_DATA_FLASH_BLOCK_SIZE);   
            if ( result != PFDL_OK) {
                seq = PFDS_WERASE;
            }
            else {
                seq = PFDS_WRITE;
            }
            break;

        case PFDS_WERASE:
            result = pfdl_erase_block(blockCounter);
			if( result != PFDL_OK ){
				Serial.print("DataFlash Erase error in PFDS_WERASE : ");
				Serial.println(result);
                seq = PFDS_ERR;
			}
            else {
                seq = PFDS_WRITE;
            }
            break;

        case PFDS_WRITE:

            result = pfdl_write(flashAddress, wrBuffa, PFDL_DATA_FLASH_BLOCK_SIZE);
			if( result != PFDL_OK ){
				Serial.print("DataFlash Write error in PFDS_WRITE : ");
				Serial.println(result);
                seq = PFDS_ERR;
			}
            else {
                seq = PFDS_WCONT;
            } 
            break;

        case PFDS_WCONT:
            blockCounter += 1;
            if (blockCounter<blockEnd) {
                flashAddress += PFDL_DATA_FLASH_BLOCK_SIZE;
                wrBuffa += PFDL_DATA_FLASH_BLOCK_SIZE;
                seq = PFDS_WBLANK;
            }
            else {
                seq = PFDS_WCLOSE;
            }
            break;

        case PFDS_WCLOSE:
            pfdl_close();
            seq = PFDS_USABLE;
            endFlag = true;
            break;

        case PFDS_ERR:
            pfdl_close();
            endFlag = true;
            break;

        default:
            Serial.print("Unexpected Sequence state error in eeprom_update_page");
            endFlag = true;
            break;
        } 
    } while (endFlag == false);
}
/**
 * @fn void eeprom_read_pages(uint8_t* buffa, uint8_t pageNo)
 * @brief FlashDataRomの指定したページの読み出しを行う
 * @param [in] buffa FlashDataRom から読み出したデータを保存するbuffa
 * @param [in] pageNo page numberを指定する
 * @note ソースに関係なく、プログラムサイズが4Kの時は4K全部を loadしている
 * @attention Multi Block (ソースに1Kbyte以外 2K, 4K, 8K) 指定の時だけ有効な関数
*/
void eeprom_read_pages(uint8_t* buffa, uint8_t pageNo)
{
    sdlSeq seqR = PFDS_USABLE;           //! シーケンス変数

    bool endFlagR = false;               //! シーケンス終了フラグ
    int  blockStartR = 0;                //! 読み込み開始ブロックの番号
    int  blockCounterR = 0;              //! Block Conter　読み込み中のブロックの番号
    int  blockEndR = 0;                  //! 読み込み終了ブロックの番号
    uint16_t flashAddressR = 0;          //! pfdl_read() 読み出し開始アドレス
    int result=0;

    do {
        switch (seqR)
        {
        case PFDS_USABLE:
             seqR = PFDS_SET;
            break;

        case PFDS_SET:          
            blockStartR = pageNo * EEPROM_PER_BLOCK;    // 開始ブロックを設定
            blockEndR = blockStartR + EEPROM_PER_BLOCK;  // 終了
            blockCounterR = blockStartR;
            seqR = PFDS_RADDR;
            break; 

        case PFDS_RADDR:
            flashAddressR = blockStartR * PFDL_DATA_FLASH_BLOCK_SIZE;
            seqR = PFDS_ROPEN;
            break;

        case PFDS_ROPEN:
            result = pfdl_open();
            if (result == PFDL_OK) {
                seqR = PFDS_READ;
            } 
            else {
                Serial.print("DataFlash Read error");
                Serial.print(result);                
                seqR = PFDS_ERR;
            }
            break;

        case PFDS_READ:
            result = pfdl_read( (uint16_t)flashAddressR, buffa, PFDL_DATA_FLASH_BLOCK_SIZE );   
            if (result != PFDL_OK) {
                Serial.print("DataFlash Read error");
                Serial.print(result);
                seqR = PFDS_ERR;
            }
            else {
                seqR = PFDS_RCONT;
            }
            break;

        case PFDS_RCONT:
            blockCounterR += 1;
            if (blockCounterR < blockEndR) {
                flashAddressR += PFDL_DATA_FLASH_BLOCK_SIZE;
                buffa+=PFDL_DATA_FLASH_BLOCK_SIZE;
                seqR = PFDS_READ;
            }
            else {
                seqR = PFDS_RCLOSE;
            }       
            break;

        case PFDS_RCLOSE:
            pfdl_close();
            seqR = PFDS_SET;
            endFlagR = true;
            break;

        case PFDS_ERR:
            pfdl_close();
            endFlagR = true;
            break;

        default:
            Serial.print("Unexpected Sequence state error in eeprom_read_pages");
            endFlagR = true;
            break;
        }
    } while (endFlagR == false);
}

/**
 * @fn void eeprom_read_topline(uint8_t pageNo)
 * @brief 指定されたページのソースを1行読み出す。
 * @details FILES [pageno] が実行された時に呼び出される。
 * @param [in] buffa  指定ページの1行目を格納するバッファ
 * @param [in] pageNum ページ指定
 * @attention Multi Block (ソースに1Kbyte以外 2K, 4K, 8K) 指定の時だけ有効な関数
 */
void eeprom_read_topline(uint8_t* buffa, uint8_t pageNum)
{
    pfdl_status_t	result = PFDL_OK;
    uint16_t topAddress = pageNum * EEPROM_PAGE_SIZE;  //! データフラッシュの指定されたページのトップアドレス

    result = pfdl_open();
    if (result == PFDL_OK) {
        result = pfdl_read( (uint16_t)topAddress, buffa, SIZE_LINE );   
        if (result != PFDL_OK) {
            Serial.print("DataFlash Read error");
            Serial.print(result);
        }
    }
    else {
        Serial.print("DataFlash Open Error:");
        Serial.print(result);
    }
    pfdl_close();
}

/**
 * @fn void eeprom_erase_pages(uint8_t pageNo)
 * @brief 指定されたページの該当する FlashDataRomのブロックを消去する
 * @param [in] pageNo 消去対象のページ番号
 * @attention Multi Block (ソースに1Kbyte以外 2K, 4K, 8K) 指定の時だけ有効な関数
 */
void eeprom_erase_pages(uint8_t pageNo)
{
    uint8_t buffa[EEPROM_PAGE_SIZE];

    for (int i=0; i<EEPROM_PAGE_SIZE; i++) {
        buffa[i] = 0;
    }
    eeprom_update_pages(buffa, pageNo);
}

// EEPROMへのRead/Write Routine 
//  RL78でArduino依存関数の代替関数
//      eeprom_read_block((void *)listbuf, (void *)topAddr, SIZE_LIST);    // ブロックリード
//      eeprom_update_block((void *)listbuf, (void *)topAddr, SIZE_LIST);  // ブロックライト
//      eeprom_update_dword(topAddr,0);                                    // 2Byte(16bit)更新

#if USE_HS_EEPROM == 1
/**
 * @fn void eeprom_read_block(uint8_t *buffa, uint16_t romAddr, size_t n)
 * @brief 内部データフラッシュへのRead Routine (RL78でArduino依存関数の代替関数)
 */
void eeprom_read_block(uint8_t *buffa, uint16_t romAddr, size_t n)
{
    pfdl_status_t	result = PFDL_OK;

    result = pfdl_open();
    if (result == PFDL_OK) {
        result = pfdl_read( (uint16_t)romAddr, buffa, n );   
        if (result != PFDL_OK) {
            Serial.print("DataFlash Read error");
            Serial.print(result);
        }
    }
    else {
        Serial.print("DataFlash Open Error:");
        Serial.print(result);
    }
    pfdl_close();
}

// 書き込みを行うエリアは erase されている必要があるので、先に blank_checkを行う
// blankでなければ eraseを行うが、1ブロック全部eraseされるので、その対応がしてある。
// PFDL_DATA_FLASH_BLOCK_SIZE = 1024 from data_flash_util.h
//
/**
 * @fn void eeprom_update_block(uint8_t* buffa, uint16_t romAddr, size_t size)
 * @brief ブロック単位の書き込み 
 * @param [in] buffa   書き込みデータバッファのポインタ
 * @param [in] romAddr 書き込み開始アドレス。ブロック開始アドレス以外も指定可能
 * @param [in] size    ブロック境界を超えない値を指定可能
 * @details 1ブロックの範囲でデータの書き込みができる
 * @warning ブロックの途中のアドレスも指定可能だが、ブロックをまたいだ書き込みはできない
 * @note ソースのLoad/SaveをPage単位にしたので使ってないが、将来の使用を考えて残してある。
 */
void eeprom_update_block(uint8_t* buffa, uint16_t romAddr, size_t size)
{
	pfdl_status_t	result = PFDL_OK;   //! pfdl_ コマンドの実行結果を格納
    uint8_t  eeprom_buffa[PFDL_DATA_FLASH_BLOCK_SIZE];  //! 1ブロック分のデータバッファ
	
 	result = pfdl_open();
	if ( result == PFDL_OK ){
		result = pfdl_blank_check( (uint16_t)romAddr, size);   
        if ( result != PFDL_OK) {
            int blockNo = romAddr / PFDL_DATA_FLASH_BLOCK_SIZE ;
            int16_t specified_block_address = PFDL_DATA_FLASH_BLOCK_SIZE * blockNo;

            // 書き込み対象のブロックのデータ（1024バイト）を eeprom_bufferに読み出す
            result = pfdl_read( specified_block_address, eeprom_buffa, PFDL_DATA_FLASH_BLOCK_SIZE); 
            if ( result != PFDL_OK) {
				Serial.print("DataFlash Read error during write Operation :");
				Serial.println(result);                
            }
            else {
                // eerpm_buffa の書き込み対象のデータを書き換える
                int16_t offset = romAddr - PFDL_DATA_FLASH_BLOCK_SIZE * blockNo;
                for (int i=0; i<size; i++) {
                    eeprom_buffa[offset + i] = *buffa++;
                }
                // 書き込み対象のブロック全体を消去する
                result = pfdl_erase_block(blockNo);
				if( result != PFDL_OK ){
					Serial.print("DataFlash Erase error :");
					Serial.println(result);
				}
				else{
                    // データの更新された eeprom_buffa をデータフラッシュに書き込む
                    result = pfdl_write(specified_block_address, eeprom_buffa, PFDL_DATA_FLASH_BLOCK_SIZE);
					if( result != PFDL_OK ){
						Serial.print("DataFlash Write error :");
						Serial.println(result);
					}                    
                }
            }       
        }
        else {      
            // 書き込み対象エリアがブランクだったときの書き込み処理
            result = pfdl_write((uint16_t)romAddr, buffa, size);
            if (result != PFDL_OK) {
				Serial.print("DataFlash Write error :");
				Serial.println(result);                
            }
        }
    }
	else{
		Serial.print("DataFlash Open error :");
		Serial.println(result);
	}
	pfdl_close(); 

}
/**
 * @fn void eeprom_clear_block( uint16_t topAddress) 
 * @brief topAddressで指定された　FlashDataRom の該当するブロックに 0 を書き込む
 * @param [in] topAddress FlashDataRom の消去を行うブロックのアドレス
 * @note  アドレスで指定しているが、該当する1ブロック全部に0を書き込む
 */
void eeprom_clear_block( uint16_t topAddress) 
{
 	pfdl_status_t result = pfdl_open();

	if ( result == PFDL_OK ){
        uint8_t blockdata[PFDL_DATA_FLASH_BLOCK_SIZE];

        for (int i=0; i<PFDL_DATA_FLASH_BLOCK_SIZE; i++)
        {
            blockdata[i] = 0;
        }
        result = result = pfdl_write(topAddress, blockdata, PFDL_DATA_FLASH_BLOCK_SIZE);
		if( result != PFDL_OK ){
			Serial.print("DataFlash clear error :");
			Serial.println(result);
		}
    }
    else {
		Serial.print("DataFlash Open error :");
		Serial.println(result);        
    }    
    pfdl_close();
}

#else
void eeprom_read_block(uint8_t *buffa, uint16_t romAddr, size_t n)
{
    for (int i=0; i<n; i++) {
        *buffa++ = EEPROM.read(romAddr++);
    }
}

void eeprom_update_block(uint8_t* buffa, uint16_t romAddr, size_t n)
{
    for (int i=0; i<n; i++) {
        EEPROM.write(romAddr++ , *buffa++);
    }
}

void eeprom_update_byte(uint16_t romAddr, uint8_t byteData)
{
    EEPROM.write( romAddr , byteData);
}
#endif

#endif
