/**
 * @file getparam.cpp
 * @author your name (you@domain.com)
 * @brief パラメータを取得する関数群
 * @version 0.1
 * @date 2021-11-22
 * @copyright Copyright (c) 2021 
 */

#include "Arduino.h"
#include "basic.h"
#include "sub.h"
#include "intercode.h"
#include "table.h"
#include "errcode.h"
#include "console.h"
#include "gpio.h"
#include "charinout.h"
#include "getparam.h"

/**
 * @brief キー入力文字コードの取得 INKEY()関数
 * @return int16_t キー入力コードを返す
 */
int16_t iinkey() {
  int16_t rc = 0;
   if (checkOpen()||checkClose()) return 0;  
  if (prevPressKey) {
    // 一時バッファに入力済キーがあればそれを使う
    rc = prevPressKey;
    prevPressKey = 0;
  } else if (c_kbhit()) {
    // キー入力
    rc = c_getch();
  }
  return rc;
}

/**
 * @brief プログラム領域空きチェック
 * @return 戻り値 プログラムの空き領域
 * @note 外部参照なし 多分変更不要
 */
int16_t getsize(void) {
  uint8_t* lp; //ポインタ
  for (lp = listbuf; *lp; lp += *lp); //ポインタをリストの末尾へ移動
  return listbuf + SIZE_LIST - lp - 1; //残りを計算して持ち帰る
}

 /**
  * @brief 演算結果等の値を取得する
  * @return int16_t 取得した値
  */
int16_t ivalue(void) {
  int16_t value; // 値

  switch (*cip++) {    // 中間コードで分岐
  case I_NUM:          // 定数の場合
  case I_HEXNUM:       // 16進定数
  case I_BINNUM:       // 2進数定数  
    value = *cip | *(cip + 1) << 8; // 定数を取得
    cip += 2;    // 中間コードポインタを定数の次へ進める
    break;

  case I_PLUS:  value = ivalue();   break;  //「+」の場合   
  case I_MINUS: value = - ivalue(); break;  //「-」の場合   
  case I_LNOT:  value = !ivalue();  break;  //「!」NOT演算
  case I_BITREV:             // 「~」 ビット反転
    value = ~((uint16_t)ivalue()); //値を取得してNOT演算
    break;

   case I_VAR:               // 変数の場合
    value = var[*cip++];     // 変数番号から変数の値を取得して次を指し示す
    break;

  case I_OPEN:               //「(」の場合
    cip--;
    value = getparamValue();      // 括弧の値を取得
    break;

  case I_ARRAY:               // 配列の場合
    value = getparamValue();       // 括弧の値を取得
    if (err)                  // もしエラーが生じたら
      break;
    if (value >= SIZE_ARRY) { // もし添え字の上限を超えたら
      err = ERR_SOR;          // エラー番号をセット
      break;
    }
    value = arr[value];       // 配列の値を取得
    break;

  // 関数の値の取得
  case I_RND:                 // 関数RNDの場合
    value = getparamValue();       // 括弧の値を取得
    if (err)                  // もしエラーが生じたら
      break;
    value = random(value);    // 乱数を取得
    break;

  case I_ABS:                 // 関数ABSの場合
    value = getparamValue();       // 括弧の値を取得
    if (value == -32768)
      err = ERR_VOF;
    if (err)
      break;
    if (value < 0) 
      value *= -1;            // 正負を反転
    break;

  case I_SIZE:                // 関数FREEの場合
    if (checkOpen()||checkClose()) break;
    value = getsize();        // プログラム保存領域の空きを取得
    break;

  case I_INKEY:   value = iinkey();   break; // 関数INKEY    
  case I_BYTE:    value = iwlen();    break; // 関数BYTE(文字列)   
  case I_LEN:     value = iwlen(1);   break; // 関数LEN(文字列)
  case I_ASC:     value = iasc();     break; // 関数ASC(文字列)
  case I_PEEK:    value = ipeek();    break; // PEEK()関数
  case I_I2CW:    value = ii2crw(1);  break; // I2CW()関数
  case I_I2CR:    value = ii2crw(0);  break; // I2CR()関数
  case I_SHIFTIN: value = ishiftIn(); break; // SHIFTIN()関数
  case I_PULSEIN: value = ipulseIn(); break; // PLUSEIN()関数
  case I_MAP:     value = imap();     break; // 関数MAP(V,L1,H1,L2,H2)
#if USE_IR == 1
  case I_IR:      value = iir();      break; // IR(ピン番号)
#endif
#if USE_GRADE == 1
  case I_GRADE:   value = igrade();   break; // 関数GRADE(値,配列番号,配列データ数)
#endif
  case I_TICK: // 関数TICK()
    if ((*cip == I_OPEN) && (*(cip + 1) == I_CLOSE)) {
      // 引数無し
      value = 0;
      cip+=2;
    } else {
      value = getparamValue(); // 括弧の値を取得
      if (err)
        break;
    }
    if(value == 0) {
        value = (millis()) & 0x7FFF;            // 0～32767msec(0～32767)
    } else if (value == 1) {
        value = (millis()/1000) & 0x7FFF;       // 0～32767sec(0～32767)
    } else {
      value = 0;                                // 引数が正しくない
      err = ERR_VALUE;
    }
    break;
      
  case I_DIN: value = iIN();  break;  // DIN(ピン番号)
  case I_ANA: value = iana(); break;  // ANA(ピン番号)
#if USE_MISAKIFONT != 0
  case I_GETFONT:value = igetfont(); break; // 美咲フォントの取得
#endif
#if USE_NEOPIXEL == 1
  case I_RGB:    value = iRGB();    break;  // RGBコード変換
  case I_NPOINT: value = inpoint(); break;  // NPOINT(X,Y)
#endif

  default: //以上のいずれにも該当しなかった場合
    cip--;
    // 仮想アドレス
    if (*cip >= I_MVAR && *cip <= I_MEM2) {
       value = pgm_read_word(vatable + *cip - I_MVAR);
       cip++;
    } else
    
    // 定数
    if (*cip >= I_OUTPUT && *cip <= I_LED) {
       value = pgm_read_byte(constValue + *cip-I_OUTPUT);
       cip++;
    } else
#if USE_ANADEF == 1
 #ifdef ARDUINO_AVR_MEGA2560
    if (*cip >= I_A0 && *cip <=I_A15) {
       value = 54 +  (*cip-I_A0);
 #else
    if (*cip >= I_A0 && *cip <=I_A7) {
       value = 14 +  (*cip-I_A0);
 #endif
       cip++;
    } else
#endif
    {
      err = ERR_SYNTAX; //エラー番号をセット
    }
  }
  return value; //取得した値を持ち帰る
}
 
/**
 * @brief 2因子演算 (* / % << >>  & | ^)
 * @return int16_t 2因子演算の結果
 */
int16_t imul(void) {
  int16_t value, tmp; // 値と演算値
  value = ivalue();   // 値を取得
  if (err)         
    return -1;
       
  for (;;) { //無限に繰り返す    
    uint8_t code = *cip;
    if(code >= I_MUL && code <= I_XOR) {
      cip++;  
      tmp = ivalue(); // 演算値を取得 
      if (code == I_MUL) {           // 掛け算 
        value *= tmp;     
      } else if (code == I_DIV) {    // 割り算
        if (tmp == 0) {
          err = ERR_DIVBY0;
        } else {
          value /= tmp;   // 割り算を実行
        }
      } else if (code == I_DIVR) {   // 商余
        if (tmp == 0) {
          err = ERR_DIVBY0;
        } else {
          value %= tmp;
        }
      } else if (code == I_LSHIFT) { // << ビット左シフト
        value =((uint16_t)value)<<tmp;       
      } else if (code == I_RSHIFT) { // >> ビット右シフト
        value =((uint16_t)value)>>tmp;  
      } else if (code == I_AND) {    // & ビットAND
        value =((uint16_t)value)&((uint16_t)tmp);        
      } else if (code == I_OR) {     // | ビットOR
        value =((uint16_t)value)|((uint16_t)tmp);        
      } else if (code == I_XOR) {    // ^ ビットXOR
        value =((uint16_t)value)^((uint16_t)tmp);        
      }
    } else {
      return value; // 値を持ち帰る      
    }
    if (err)
      return -1;
  }
}

/**
 * @brief 加算か減算の計算
 * @return int16_t 　計算結果
 */
int16_t iplus(void) {
  int16_t imul();
  int16_t value, tmp; // 値と演算値

  value = imul(); // 値を取得
  if (err)
    return -1;

  for (;;) // 無限に繰り返す
  switch(*cip){   // 中間コードで分岐
  case I_PLUS:    // 足し算の場合
    cip++;        // 中間コードポインタを次へ進める
    tmp = imul(); // 演算値を取得
    value += tmp; // 足し算を実行
    break;

  case I_MINUS:   // 引き算の場合
    cip++;        // 中間コードポインタを次へ進める
    tmp = imul(); // 演算値を取得
    value -= tmp; // 引き算を実行
    break;

  default: // 以上のいずれにも該当しなかった場合
    return value; 
  }
}

/**
 * @brief The Parser
 * @return int16_t 取得した値
 */
int16_t iexp(void) {
  int16_t value, tmp; // 値と演算値
  value = iplus();    // 値を取得
  if (err)
    return -1; 

  // conditional expression 
  for (;;) { // 無限に繰り返す
    uint8_t code = *cip;
    if (code >= I_EQ && code <= I_LOR) {
      cip++;
      tmp = iplus(); // 演算値を取得
      if (code == I_EQ) {        //「=」の場合
        value = (value == tmp);
      } else if (code == I_NEQ || code == I_NEQ2) { //「!=」「<>」の場合
        value = (value != tmp);
      } else if (code == I_LT) {  //「<」の場合
        value = (value < tmp);
      } else if (code == I_LTE) { //「<=」の場合
        value = (value <= tmp);
      } else if (code == I_GT)  { //「>」の場合
        value = (value > tmp);
      } else if (code == I_GTE) { //「>=」の場合
        value = (value >= tmp);
      } else if (code == I_LAND){ // AND (論理積)
        value = (value && tmp);
      } else if (code == I_LOR){  // OR (論理和)
        value = (value || tmp);    
      }
    } else {
      break;
    }
  }
  return value;    
}

//*** コマンド・関数引数評価 ****************

// コマンド引数評価(符号付き整数の取得, 範囲チェックあり)
// 戻り値 評価した引数の値（エラー時はerrにエラーコードをセットする）
uint8_t getParamRange(int16_t *prm, int16_t v_min, int16_t v_max, uint8_t flgCmma) {
  *prm = iexp(); 
  if (!err &&  (*prm < v_min || *prm > v_max)) 
    err = ERR_VALUE;
  else if (flgCmma && *cip++ != I_COMMA) {
    err = ERR_SYNTAX;
 }
  return err;
}

// コマンド引数取得(符号付き整数の取得, 引数チェックなし)
// 戻り値 評価した引数の値（エラー時はerrにエラーコードをセットする）
uint8_t getParamInt16(int16_t *prm, uint8_t flgCmma) {
  *prm = iexp(); 
  if (!err && flgCmma && *cip++ != I_COMMA) {
   err = ERR_SYNTAX;
  }
  return err;
}

// コマンド引数取得(符号無し整数の取得, 引数チェックなし)
// 戻り値 評価した引数の値（エラー時はerrにエラーコードをセットする）
uint8_t getParamUint16(uint16_t* prm, uint8_t flgCmma) {
  *prm = iexp(); 
  if (!err && flgCmma && *cip++ != I_COMMA) {
   err = ERR_SYNTAX;
  }
  return err;
}

// コマンド引数取得(符号付き32ビット整数の取得, 引数チェックなし)
// 戻り値 評価した引数の値（エラー時はerrにエラーコードをセットする）
// ※この関数は、符号付き引数の範囲が-32768 ～ 32767を超える場合に利用
uint8_t getParamInt32(int32_t* prm, uint8_t flgCmma) {
  *prm = iexp(); 
  if (!err && flgCmma && *cip++ != I_COMMA) {
   err = ERR_SYNTAX;
  }
  return err;
}

// 単一引数の評価（配列の添え字評価等に利用
// 例： "(" 値 ")" の処理
int16_t getparamValue() {
  int16_t value; // 値
  if ( checkOpen() || getParamInt16(&value, false) || checkClose() )
    return 0;
  return value;
}
