/** @file gpio.cpp
 * @brief I/O pin 処理
 * @todo 2020/07/19 ilprint()が未実装 iprint() in basic.cppの修正が必要
 * @todo 2020/08/07 GR-ADZUKI対応のピン機能チェックテーブルの検証がしていない
 * @todo 2020/08/07 IsUseablePin に範囲チェックの追加 割り当てのないピンの返り値が未定義
 */
#include "Arduino.h"
#include "basic.h"
#include "sub.h"
#include "charinout.h"
#include "IR.h"
//#include "console.h"
#include "intercode.h"
#include "errcode.h"
#include "getparam.h"
#include "iprint.h"
#include "ttconfig.h"
#include "gpio.h"

#define IsPWM_PIN(N) IsUseablePin(N,FNC_PWM)      // 指定ピンPWM利用可能判定
#define IsADC_PIN(N) IsUseablePin(N,FNC_ANALOG)   // 指定ピンADC利用可能判定
#define IsIO_PIN(N)  IsUseablePin(N,FNC_IN_OUT)   // 指定ピンデジタル入出力利用可能判定

// ピン機能利用設定フラグ(ORによる重複指定OK)
#define FNC_IN_OUT  1  // デジタルIN/OUT
#define FNC_PWM     2  // PWM
#define FNC_ANALOG  4  // アナログIN

enum {
    RTC_YEAR = 0,
    RTC_MONTH,
    RTC_DAY,
    RTC_WEEK,
    RTC_HOUR,
    RTC_MINUTE,
    RTC_SECOND
};

// GR-ADZUKI ピン機能チェックテーブル
//  @note Arduino uno 3と同じにしたが、 20, 21 は未定義のはずだが？
#if defined(GRADZUKI)
const PROGMEM uint8_t pinFunc[] = {
  0,0,1,1|2,1,1|2,1|2,1,1,1|2,          // ポート0 -  9: 
  1|2,1|2,1,1,1|4,1|4,1|4,1|4,1|4,1|4,  // ポート10 - 19: 
  1|4,1|4,
};

#else
  #error "Board Undefined "
#endif

/**
 * @brief ピン利用可能チェック
 * @param pinno ピン番号
 * @param fnc 
 * @return uint8_t 
 * @note どこからも呼ばれてない
 */
inline uint8_t IsUseablePin(uint8_t pinno, uint8_t fnc) {
  return pgm_read_byte_near(&pinFunc[pinno]) & fnc;
}

// **** I2Cライブラリの利用設定(オプション) ********
#if USE_CMD_I2C == 1
//  #include <Wire.h>
    #include "wire.h"
#endif 

#if USE_SO1602AWWB == 1
//*** I2C接続キャラクタディスプレイ **************
//#include "src/lib/SO1602AWWB.h"
#include "SO1602AWWB.h"
#endif 

/**
 * @brief GPIO ピン機能設定
 */
void igpio() {
  int16_t pinno;  // ピン番号
  int16_t pmode;  // 入出力モード

  // 入出力ピンの指定
  if ( getParamRange(&pinno, 0, TT_MAX_PINNUM, true) || // ピン番号取得
       getParamRange(&pmode, 0, 2, false) ) return;     // 入出力モードの取得

  // デジタル入出力として利用可能かチェック
  if (!IsIO_PIN(pinno)) {
    err = ERR_GPIO;
    return;    
  }
  pinMode(pinno, pmode);
}


/**
 * @brief GPIO ピンデジタル出力
 */
void iout() {
  int16_t pinno,  data;

  if ( getParamRange(&pinno, 0, TT_MAX_PINNUM, true) || // ピン番号取得
       getParamInt16(&data, false) ) return;            // データ指定取得

  data = data ? HIGH: LOW;
  if (! IsIO_PIN(pinno)) {
    err = ERR_GPIO;
    return;
  }
  pinMode(pinno, OUTPUT);     // ピンモード設定
  digitalWrite(pinno, data);  // データ出力
}

/**
 * @brief LEDピンデジタル出力
 */
void iled() {  
  int16_t pinno,  data;

  if ( getParamInt16(&data, false) ) return;            // データ指定取得
  data = data ? HIGH: LOW; 

  // データ出力
  pinMode(LED_BUILTIN,OUTPUT);
  digitalWrite(LED_BUILTIN, data);
}

/**
 * @brief GPIO ピンデジタル入力
 * @return int16_t デジタル入力データ
 * @note IN(ピン番号[,Float|PullUp)
 * @todo if( checkOpen() ... のreturn値が未定義だったので、仮に0にしたが未確認
 */
int16_t iIN() {
  int16_t pinNo, mode = INPUT;
  int16_t value = 0;

  if ( checkOpen()|| getParamRange(&pinNo,0,TT_MAX_PINNUM, false) ) 
  {
    err = ERR_GPIO;             // 未検証のエラー
    return 0; // ピン番号取得  何も返してないので0にしてエラーにした。
  }
  if (*cip == I_COMMA) {
    cip++;
    if (getParamInt16(&mode, false)) return 0;
    if ( (mode !=INPUT) && (mode != INPUT_PULLUP) ) {
      err = ERR_VALUE;
      return 0;
    }    
  }
  if (checkClose()) return 0;
  
  if ( !IsIO_PIN(pinNo) )  {
    err = ERR_GPIO;
  } else {
    pinMode(pinNo,mode);
    value = digitalRead(pinNo);  // 入力値取得
  }
  return value;
}
 
/**
 * @brief GPIO アナログ入力
 * @return int16_t アナログ入力値
 * @note Usage: ANA(ピン番号)
 */
int16_t iana() {
  int16_t value;
  if ( !(checkOpen()|| getParamRange(&value,0,TT_MAX_PINNUM, false) || checkClose()) ) {
    if ( !IsADC_PIN(value) ) {
      err = ERR_GPIO;
    } else {
      value = analogRead(value);    // 入力値取得
    }
  }
  return value; 
}

/**
 * @brief PWMコマンド
 * @note Usage: PWM ピン番号, DutyCycle(0~255)
 */
void ipwm() {
  int16_t pinno;      // ピン番号
  int16_t duty;       // デューティー値 0～255

  if ( getParamRange(&pinno, 0, TT_MAX_PINNUM, true) ||    // ピン番号取得
       getParamRange(&duty,  0, 255, false) ) return;      // デューティー値

  // PWMピンとして利用可能かチェック
  if (!IsPWM_PIN(pinno)) {
    err = ERR_GPIO;
    return;    
  }
  analogWrite(pinno, duty);
}


/**
 * @brief shiftOutコマンド SHIFTOUT dataPin, clockPin, bitOrder, value
 */
void ishiftOut() {
#if USE_SHIFTOUT == 1
  int16_t dataPin, clockPin;
  int16_t bitOrder;
  int16_t data;

  if (getParamRange(&dataPin, 0,TT_MAX_PINNUM, true)) return;
  if (getParamRange(&clockPin,0,TT_MAX_PINNUM, true)) return;
  if (getParamRange(&bitOrder,0,1, true)) return;
  if (getParamRange(&data, 0,255,false)) return;

  if ( !IsIO_PIN(dataPin) ||  !IsIO_PIN(clockPin) ) {
    err = ERR_GPIO;
    return;
  }
  pinMode(dataPin,OUTPUT);
  pinMode(clockPin,OUTPUT);
  shiftOut(dataPin, clockPin, bitOrder, data);
#endif
}

#if USE_CMD_I2C == 1
/**
 * @brief I2C 初期化
 * @todo 要チェック どこからも呼ばれてないが、Wire.begin()でマスターとして初期化のはず
 */
void i2c_init() {
  Wire.begin();                      // I2C利用開始 
}
#endif


/**
 * @brief I2Cデータ送受信
 * @param mode モード設定
 * @note  mode 1: I2CW関数  I2CW(I2Cアドレス, コマンドアドレス, コマンドサイズ, データアドレス, データサイズ)
 * @note  mode 0: I2CR関数  I2CR(I2Cアドレス, コマンドアドレス, コマンドサイズ, 受信データアドレス,受信データサイズ)
 * @return int16_t I2C通信結果
 */
int16_t ii2crw(uint8_t mode) {
#if USE_CMD_I2C == 1
  int16_t  i2cAdr, ctop, clen, top, len;
  uint8_t* ptr;
  uint8_t* cptr;
  int16_t  rc;

  if (checkOpen() ||
      getParamRange(&i2cAdr, 0,  0x7f, true) ||
      getParamRange(&ctop,   0, 32767, true) ||
      getParamRange(&clen,   0, 32767, true) ||
      getParamRange(&top,    0, 32767, true) ||
      getParamRange(&len,    0, 32767,false) ||
      checkClose()
   ) return 0;

  ptr  = v2realAddr(top);
  cptr = v2realAddr(ctop);
  if (ptr == 0 || cptr == 0 || v2realAddr(top+len) == 0 || v2realAddr(ctop+clen) == 0) 
     { err = ERR_VALUE; return 0; }

  if (mode) {
  // I2Cデータ送信
    Wire.beginTransmission(i2cAdr);
    if (clen) Wire.write(cptr, clen);
    if (len)  Wire.write(ptr, len);
    rc =  Wire.endTransmission();
    return rc;
  } else {
    // I2Cデータ送受信
    Wire.beginTransmission(i2cAdr);
    if (clen) Wire.write(cptr, clen);
    rc = Wire.endTransmission();
    if (len) {
      if (rc!=0)
        return rc;
      Wire.requestFrom(i2cAdr, len);
      while (Wire.available()) {
        *(ptr++) = Wire.read();
      }
    }  
    return rc;    
  }
#else
  return 1;
#endif
}

#if USE_SHIFTIN ==1
/**
 * @brief SHIFT IN 動作を行う
 * @param ulDataPin I2C データピン番号
 * @param ulClockPin I2C クロックピン番号
 * @param ulBitOrder ビット並びの設定
 * @param lgc 　　クロックピンのロジックレベル
 * @return uint8_t 読み出したロジックレベル
 */
uint8_t _shiftIn( uint8_t ulDataPin, uint8_t ulClockPin, uint8_t ulBitOrder, uint8_t lgc){
  uint8_t value = 0 ;
  uint8_t i ;
  for ( i=0 ; i < 8 ; ++i ) {
    digitalWrite( ulClockPin, lgc ) ;
    if ( ulBitOrder == LSBFIRST )  value |= digitalRead( ulDataPin ) << i ;
    else  value |= digitalRead( ulDataPin ) << (7 - i) ;
    digitalWrite( ulClockPin, !lgc ) ;
  }
  return value ;
}
#endif

/**
 * @brief SHIFTIN関数 SHIFTIN(データピン, ピンモード, クロックピン, オーダ[,ロジック])
 * @note ピンモードは、FLOAT(0) or PULLUP(2)
 * @return int16_t 
 */
int16_t ishiftIn() {
#if USE_SHIFTIN == 1
  int16_t rc;
  int16_t dataPin, clockPin;
  int16_t mode,bitOrder;
  int16_t lgc = HIGH;

  if (checkOpen() ||
      getParamRange(&dataPin , 0,TT_MAX_PINNUM, true) ||  // データピン
      getParamInt16(&mode, true) ||                       // ピンモード
      getParamRange(&clockPin,0,TT_MAX_PINNUM, true) ||   // クロックピン
      getParamRange(&bitOrder,0,1, false)                 // オーダー
   ) return 0;
 
  if (*cip == I_COMMA) {
    cip++;
    if (getParamRange(&lgc,LOW, HIGH, false)) return 0;
  }
  if (checkClose()) return 0;
  
  // ピン利用の有効性チェック
  if ( !IsIO_PIN(dataPin) ||  !IsIO_PIN(clockPin) ) {
    err = ERR_GPIO;
    return 0;
  }
  
  // ピンモードのチェック
  if ( (mode != INPUT) && (mode != INPUT_PULLUP) ) { 
    err = ERR_VALUE;
    return 0;
  }

  // ピンモードの設定
  pinMode(dataPin,  mode);     // データピン
  pinMode(clockPin, OUTPUT);   // クロックピン

  // データの読み取り
  rc = _shiftIn((uint8_t)dataPin, (uint8_t)clockPin, (uint8_t)bitOrder, lgc);
  return rc;
#else
  return 0;
#endif
}

// PULSEIN関数 PULSEIN(入力ピン, ピンモード, 検出モード, タイムアウト時間[,単位指定])
int16_t ipulseIn() {
#if USE_PULSEIN == 1
  int32_t rc=0;
  int16_t dataPin, mode; // ピン
  int16_t sigMode;       // 信号検出モード
  int16_t tmout;         // タイムアウト時間(0:ミリ秒 1:マイクロ秒)
  int16_t scale =1;      // 戻り値単位指定 (1～32767)

  // コマンドライン引数の取得
  if (checkOpen() ||                              // '('のチェック
      getParamRange(&dataPin, 0,TT_MAX_PINNUM, true) || // ピン
      getParamInt16(&mode, true) ||                     // ピンモード
      getParamRange(&sigMode, LOW, HIGH, true) ||       // 信号検出モード
      getParamRange(&tmout,0,32767, false)              // タイムアウト時間(ミリ秒)
   ) return 0;           

  if (*cip == I_COMMA) {
    cip++;
    if (getParamRange(&scale,1, 32767, false)) return 0;        // 戻り値単位指定 (1～32767)
  }
  if (checkClose()) return 0;                             // ')'のチェック

  // ピン利用の有効性チェック
  if ( !IsIO_PIN(dataPin) ) {
    err = ERR_GPIO;
    return 0;
  }
  
  // ピンモードのチェック
  if ( (mode != INPUT) && (mode != INPUT_PULLUP) ) { 
    err = ERR_VALUE;
    return 0;
  }

  // ピンモードの設定
  pinMode(dataPin,  mode);     // データピン

  // データの読み取り
  rc = pulseIn(dataPin, sigMode, ((uint32_t)tmout)*1000)/scale;  // パルス幅計測
  if (rc > 32767) rc = -1; // オーバーフロー時は-1を返す
  return rc;
#else
  return 0;
#endif  
}

#if USE_CMD_I2C == 1 && USE_RTC_DS3231 == 1
#define BCD(c) (uint8_t)((c/10)*16+(c%10))
// SETDATEコマンド  SETDATE 年,月,日,曜日,時,分,秒
// @todo 2020/08/07 I2C RTC に間違った修正を加えているかも。要確認
void isetDate() {

    int year;
    int month;
    int day;
    int hour;
    int minute;
    int second;
    int week;
    int week_flag;

  if ( getParamRange( &year , 2000,2099, true) ) return; // 年
  if ( getParamRange( &month,    1,  12, true) ) return; // 月
  if ( getParamRange( &day,      1,  31, true) ) return; // 日
  if ( getParamRange( &week,     1,  7 , true) ) return; // 曜日の元データ
  if ( getParamRange( &hour,     0,  23, true) ) return; // 時
  if ( getParamRange( &minute,   0,  59, true) ) return; // 分
  if ( getParamRange( &second,   0,  59, false)) return; // 秒   
  
  switch(week) {
      case 1: week_flag = RTC_WEEk_SUNDAY;     break;
      case 2: week_flag = RTC_WEEK_MONDAY;     break;
      case 3: week_flag = RTC_WEEK_TUESDAY;    break;
      case 4: week_flag = RTC_WEEK_WEDNESDAY;  break;
      case 5: week_flag = RTC_WEEK_THURSDAY;  break;
      case 6: week_flag = RTC_WEEK_FRIDAY;  break;   
      case 7: week_flag = RTC_WEEK_SATURDAY;  break;          
      default: week_flag = RTC_WEEK_SUNDAY; break;
  }

  // RTCの設定
 // Wire.beginTransmission(0x68);
 // Wire.write(0x00);
    rtc.begin();

/*   for (uint8_t i=0; i<7; i++)
     Wire.write( BCD(t[i]));
  if (Wire.endTransmission())
    err = ERR_I2CDEV; */
    
}

// 該当変数に値をセット
void setValueTo(uint16_t* rcv,uint8_t n) {
  int16_t index;  
  for (uint8_t i=0; i <n; i++) {    
    if (*cip == I_VAR) {          // 変数の場合
      cip++; index = *cip;        // 変数インデックスの取得
      var[index] = rcv[i];        // 変数に格納
      cip++;
    } else if (*cip == I_ARRAY) { // 配列の場合
      cip++;
      index = getparamValue();         // 添え字の取得
      if (err) return;  
      if (index >= SIZE_ARRY || index < 0 ) {
         err = ERR_SOR;
         return; 
      }
      arr[index] = rcv[i];        // 配列に格納
    } else {
      err = ERR_SYNTAX;           // 変数・配列でない場合はエラーとする
      return;   
    }     
    if(i != n-1) {
      if (*cip != I_COMMA) {      // ','のチェック
         err = ERR_SYNTAX;
         return; 
      }
      cip++;
    }
  }
}

// RTCからのデータ取得
uint8_t readRTC(uint16_t* rcv, uint8_t pos, uint8_t num){
  Wire.beginTransmission(0x68);
  Wire.write(pos);
  if (Wire.endTransmission()) {
    err = ERR_I2CDEV;
    return 1;
  }
  Wire.requestFrom((uint8_t)0x68, num);
  uint8_t d,cnt = 0;
  while (Wire.available()) {
    d = Wire.read();
    rcv[num-1-cnt] = (d>>4)*10 + (d & 0xf);
    cnt++;
    if (cnt == num) break;
  }
  return 0;
}

// GETDATEコマンド  GETDATE 年格納変数,月格納変数, 日格納変数, 曜日格納変数
void igetDate() {
  uint16_t rcv[4];
  if (readRTC(rcv,3,4))
    return;
  rcv[0]+=2000;
  setValueTo(rcv, 4);
}

// GETDATEコマンド  SETDATE 時格納変数,分格納変数, 秒格納変数
void igetTime() {
  uint16_t rcv[3];
  if (readRTC(rcv,0,3))
    return;
  setValueTo(rcv, 3);
}

// DATE$()
void idatestr(uint8_t devno) {
  if (checkOpen()) return;
  if (checkClose()) return;
  idate(devno);    
}
#endif

// *** 有機ELキャラクタディスプレイ SO1602AWWB  ***
#if USE_SO1602AWWB == 1 && USE_CMD_I2C == 1
// OLEDキャラクタディスプレイのカーソル移動
// CLOCATE x,y
void iclocate() {
  int16_t x, y;
  if ( getParamRange(&x, 0,15, true) ) return;
  if ( getParamRange(&y, 0,1, false) ) return;

  // カーソル移動
  if (OLEDPos(x,y)) {
    err = ERR_I2CDEV;
  }
}

// OLEDキャラクタディスプレイ 画面クリア
// CCLS
void iccls() {
  if (OLEDcls())
    err = ERR_I2CDEV;
}

// OLEDキャラクタディスプレイ コントラスト設定
// CCONS n
void iccons() {
  int16_t cons;
  if ( getParamRange(&cons, 0,255, false) ) return;
  if (OLEDcontrast(cons))
    err = ERR_I2CDEV;  
}

// OLEDキャラクタディスプレイ 画面表示・非表示設定
// CDISP n
void icdisp() {
  int16_t sw;
  if ( getParamRange(&sw, 0,1, false) ) return;
  if (OLEDisplay(sw))
    err = ERR_I2CDEV;  
}

// OLEDキャラクタディスプレイ カーソル表示・非表示設定
// CCURS n
void iccurs() {
  int16_t sw;
  if ( getParamRange(&sw, 0,1, false) ) return;
  if (OLEDCurs(sw))
    err = ERR_I2CDEV;  
}

// OLEDキャラクタディスプレイ 文字列表示
// CPRINT
void icprint() {
  iprint(CDEV_CLCD, 0);
}
#endif

// *** 液晶キャラクタディスプレイ SC1602BS  ***
#if USE_SC1602BS == 1 
// 液晶キャラクタディスプレイのカーソル移動
// CLOCATE x,y
#include "SC1602BS.h"

void illocate(void) {
  int16_t x, y;
  if ( getParamRange(&x, 0,15, true) ) return;
  if ( getParamRange(&y, 0,1, false) ) return;

  // カーソル移動
  LCDPos(x,y);
}

// 液晶キャラクタディスプレイ 画面クリア
// CCLS
void ilcls(void) {
    LCDcls();
}

// 液晶キャラクタディスプレイ コントラスト設定
// CCONS n
void ilcons(void) {
  int16_t cons;
  if ( getParamRange(&cons, 0,255, false) ) return;
  LCDcontrast(cons);
}

// 液晶キャラクタディスプレイ 画面表示・非表示設定
// CDISP n
void ildisp() {
  int16_t sw;
  if ( getParamRange(&sw, 0,1, false) ) return;
  LCDisplay(sw); 
}

// 液晶キャラクタディスプレイ カーソル表示・非表示設定
// CCURS n
void ilcurs() {
  int16_t sw;
  if ( getParamRange(&sw, 0,1, false) ) return;
  LCDCurs(sw);
}

// 液晶キャラクタディスプレイ 文字列表示
// CPRINT
void ilprint() {
//  iprint(CDEV_CLCD);
}
#endif
