/**
 * @file sub.cpp
 * @brief 各種関数
 */

#include "Arduino.h"
#include "basic.h"
#include "charlib.h"
#include "console.h"
#include "charinout.h"
#include "device.h"
#include "intercode.h"
#include "table.h"
#include "errcode.h"
#include "getparam.h"
#include "sub.h"

// 美咲フォントの利用
#if USE_MISAKIFONT != 0
#include "src/lib/misakiSJIS500.h"
#endif

// 2進数の出力
// 引数
//  value : 出力対象数値
//  d     : 桁指定(0で指定無し)
//  devno : デバイス番号
// 機能
// 'BBBBBBBBBBBBBBBB' B:数値
//  dで桁指定時は0補完する
//  符号は考慮しない
// 
void putBinnum(int16_t value, uint8_t d, uint8_t devno) {
  uint16_t  bin = (uint16_t)value; // 符号なし16進数として参照利用する
  uint16_t  dig = 0;               // 先頭が1から始まる桁数

  // 最初に1が現れる桁を求める
  for (uint8_t i=0; i < 16; i++) {
    if ( (0x8000>>i) & bin ) {
      dig = 15 - i;
      break;
    }
  }
  dig++;
  
  // 実際の桁数が指定表示桁数を超える場合は、実際の桁数を採用する
  if (d > dig) 
    dig = d;

  // ビット文字列の出力処理
  for (int8_t i=dig-1; i>=0; i--)
   c_putch((bin & (1<<i)) ? '1':'0', devno);
}

// 10進数の出力
// 引数
//  value : 出力対象数値
//  d     : 桁指定(0で指定無し),負の値字0埋め
//  devno : デバイス番号
// 機能
// 'SNNNNN' S:符号 N:数値 or 空白
//  dで桁指定時は空白補完する
//
void putnum(int16_t value, int16_t d, uint8_t devno) {
  uint8_t dig;  // 桁位置
  uint8_t sign; // 負号の有無（値を絶対値に変換した印）
  uint16_t new_value;
  char str[7];
  char c = ' ';
  if (d < 0) {
    d = -d;
    c = '0';
  }

  if (value < 0) {     // もし値が0未満なら
    sign = 1;          // 負号あり
    new_value = -value;
  } else {
    sign = 0;          // 負号なし
    new_value = value;
  }

  str[6] = 0;         // 終端を置く
  dig = 6;             // 桁位置の初期値を末尾に設定
  do {
    str[--dig] = (new_value % 10) + '0';  // 1の位を文字に変換して保存
    new_value /= 10;                      // 1桁落とす
  } while (new_value > 0);                // 値が0でなければ繰り返す

  if (sign) { //もし負号ありなら
    if (c == ' ') {
      str[--dig] = '-'; // 負号を保存
    } else {
      c_putch('-',devno);
      dig--;
    }
  }
  while (6 - dig < d) { // 指定の桁数を下回っていれば繰り返す
    c_putch(c,devno);   // 桁の不足を空白で埋める
    d--;                // 指定の桁数を1減らす
  }  
  c_puts_devno(&str[dig],devno);   // 桁位置からバッファの文字列を表示
}

void putHexnum(int16_t value, uint8_t d, uint8_t devno) {
  uint16_t  hex = (uint16_t)value; // 符号なし16進数として参照利用する
  uint8_t   h;
  uint8_t   dig;
  
  // 表示に必要な桁数を求める
  if (hex >= 0x1000) 
    dig = 4;
  else if (hex >= 0x100) 
    dig = 3;
  else if (hex >= 0x10) 
    dig = 2;
  else 
    dig = 1;

  if (d != 0 && d > dig) 
    dig = d;

  for (int8_t i = dig-1; i >= 0; i--) {
    h = (hex >>(i*4)) & 0x0f;
    c_putch((h >= 0 && h <= 9) ? h + '0': h + 'A' - 10, devno);
  }
}

// 16進/2進文字出力
// HEX$(数値,桁数) or HEX$(数値)
// BIN$(数値, 桁数) or BIN$(数値)
// 引数
//  devno : 出力先デバイス
//  mode  : 0:16進数 1:2進数
void ihexbin(uint8_t devno, uint8_t mode) {
  int16_t value; // 値
  int16_t d = 0; // 桁数(0で桁数指定なし)

  if (checkOpen() || getParamInt16(&value,false) ) return;  
  if (*cip == I_COMMA) {
     cip++;
     if (getParamRange(&d,0,mode?16:4,false)) return;  
  }
  if (checkClose()) return;  
  if (mode)
    putBinnum(value, d, devno);
  else
    putHexnum(value, d, devno);
}

// CHR$() 全角対応
void ichr(uint8_t devno) {
  uint16_t value; // 値
  if (checkOpen()) return;
  for(;;) {
    if (getParamUint16(&value,false)) return;
    if (value <= 0xff) {
       c_putch(value, devno);
    } else {
       c_putch(value>>8,  devno);
       c_putch(value&0xff,devno);
    }
    if (*cip == I_COMMA) {
       cip++;
       continue;
    }
    break;
  }
  if (checkClose()) return;
}

// MAP(V,L1,H1,L2,H2)関数の処理
int16_t imap() {
  int16_t value,l1,h1,l2,h2,rc;

  if (checkOpen() || getParamInt16(&value,true)||getParamInt16(&l1,true)||
       getParamInt16(&h1,true)||getParamInt16(&l2,true)||getParamInt16(&h2,false) ) {
    return 0;
  }
  if (checkClose()) return 0;
  if (l1 >= h1 || l2 >= h2 || value < l1 || value > h1) {
    err = ERR_VALUE;
    return 0;
  }
  rc = map(value,l1,h1,l2,h2);
  return rc;  
}

#if USE_GRADE == 1
// 関数GRADE(値,配列番号,配列データ数)
int16_t igrade() {
  int16_t value,arrayIndex,len,rc=-1;

  if (checkOpen() ||
      getParamInt16(&value, true) ||
      getParamRange(&arrayIndex, 0, SIZE_ARRY-1, true) ||
      getParamRange(&len, 0, SIZE_ARRY, false) ||
      checkClose()
  ) return 0;    
 
  if ( arrayIndex+len > SIZE_ARRY ) {
    err = ERR_VALUE;
    return 0;    
  }
  for (int16_t i=0; i < len; i++) {
    if (value >= arr[arrayIndex+i])  {
      rc = i;
      break;
    }
  }
  return rc;
}
#endif  

// ASC(文字列)
// ASC(文字列,文字位置)
// ASC(変数,文字位置)
int16_t iasc() {
  uint16_t value =0;
  int16_t len;     // 文字列長
  int16_t pos =1;  // 文字位置
  int16_t index;   // 配列添え字
  uint8_t* str;    // 文字列先頭位置
  
  if (checkOpen()) return 0;
  if ( *cip == I_STR) {  // 文字列定数の場合
     cip++;  len = *cip; // 文字列長の取得
     cip++;  str = cip;  // 文字列先頭の取得
     cip+=len;
  } else if ( *cip == I_VAR) {   // 変数の場合
     cip++;   str = v2realAddr(var[*cip]);
     len = *str;
     str++;
     cip++;     
  } else if ( *cip == I_ARRAY) { // 配列変数の場合
     cip++; 
     if (getParamRange(&index, 0, SIZE_ARRY-1, false)) return 0;
     str = v2realAddr(arr[index]);
     len = *str;
     str++;
  } else {
    err = ERR_SYNTAX;
    return 0;
  }
  if ( *cip == I_COMMA) {
    cip++;
    if (getParamRange(&pos,1,len,false)) return 0;
  }

  int16_t tmpPos = 0;
  int16_t i;
  for (i = 0; i < len; i++) {      
    if (pos == tmpPos+1)
      break;
    if (isZenkaku(str[i])) {
      i++;  
    }
    tmpPos++;
  }
  if (pos != tmpPos+1) {
    value = 0;
  } else {
    value = str[i];
    if(isZenkaku(str[i])) {
      value<<=8;
      value+= str[i+1];
    }
  }
  checkClose();
  return value;

}

// WLEN(文字列)
// 全角文字列長取得
int16_t iwlen(uint8_t flgZen) {
  int16_t len = 0; // 文字列長
  int16_t index;   // 配列添え字
  uint8_t* str;    // 文字列先頭位置
  int16_t wlen = 0;
  int16_t pos = 0;
  
  if (checkOpen()) 
    return 0;
    
  if ( *cip == I_VAR)  {
    // 変数の場合
     cip++;
     str = v2realAddr(var[*cip]);
     len = *str; // 文字列長の取得
     str++;      // 文字列先頭
     cip++;     
  } else if ( *cip == I_ARRAY) {
    // 配列変数の場合
     cip++; 
     if (getParamRange(&index, 0, SIZE_ARRY-1, false)) return 0;
     str = v2realAddr(arr[index]);
     len = *str; // 文字列長の取得
     str++;      // 文字列先頭
  } else if ( *cip == I_STR) {
    // 文字列定数の場合
     cip++;  len = *cip; // 文字列長の取得
     cip++;  str = cip;  // 文字列先頭の取得
     cip+=len;
  } else {
    err = ERR_SYNTAX;
  }
  checkClose();
  if (flgZen) {
    // 文字列をスキャンし、長さを求める
    while(pos < len) {
      if (isZenkaku(*str)) {
        str++;
        pos++;  
      }
      wlen++;
      str++;
      pos++;
    }  
  } else {
    wlen = len;
  }
  return wlen;
}

#if USE_DMP == 1
// 小数点数値出力 DMP$(数値) or DMP(数値,小数部桁数) or DMP(数値,小数部桁数,整数部桁指定)
void idmp(uint8_t devno) {
  int32_t value;     // 値
  int32_t v1,v2;
  int16_t n = 2;    // 小数部桁数
  int16_t dn = 0;   // 整数部桁指定
  int32_t base=1;
  
  if (checkOpen()) return;
  if (getParamInt32(&value, false)) return;
  if (*cip == I_COMMA) { 
    cip++; 
    if (getParamRange(&n, 0,4,false)) return;
    if (*cip == I_COMMA) { 
       cip++; 
      if (getParamRange(&dn,-6,6,false)) return;
    }  
  }
  if (checkClose()) return;
  
  for (int16_t i=0; i<n;i++) {
    base*=10;
  }
  v1 = value / base;
  v2 = value % base;
  if (v1 == 0 && value <0)
    c_putch('-',devno);  
  putnum(v1, dn, devno);
  if (n) {
    c_putch('.',devno);
    putnum(v2<0 ?-v2:v2, -n, devno);
  }
}
#endif

// 文字列参照 WSTR$(変数)
// WSTR(文字列参照変数|文字列参照配列変数|文字列定数,[pos,n])
// ※変数,配列は　[LEN][文字列]への参照とする
// 引数
//  devno: 出力先デバイス番号
// 戻り値
//  なし
//
void iwstr(uint8_t devno) {
  int16_t len;  // 文字列長
  int16_t top;  // 文字取り出し位置
  int16_t n;    // 取り出し文字数
  int16_t index;
  uint8_t *ptr;  // 文字列先頭
  
  if (checkOpen()) return;
  if (*cip == I_VAR) {
    // 変数
    cip++;
    ptr = v2realAddr(var[*cip]);
    len = *ptr;
    ptr++;
    cip++;
  } else if (*cip == I_ARRAY) {
    // 配列変数
    cip++; 
    if (getParamRange(&index, 0, SIZE_ARRY-1, false)) return;
    ptr = v2realAddr(arr[index]);
    len = *ptr;
    ptr++;    
  } else if (*cip == I_STR) {
    // 文字列定数
    cip++;
    len = *cip;
    cip++;
    ptr = cip;
    cip+=len;    
  } else {
    err = ERR_SYNTAX;
    return;
  }
  top = 1; // 文字取り出し位置
  n = len; // 取り出し文字数
  if (*cip == I_COMMA) {
    // 引数：文字取り出し位置、取り出し文字数の取得 
    cip++;
    if ( getParamRange(&top, 1,len,true) ||
         getParamRange(&n,1,len-top+1,false)
    ) return;
  }
  if (checkClose()) return;

  // 全角を考慮した文字位置の取得
  int16_t i;
  int16_t wtop = 1;
  for (i=0; i < len; i++) {
    if (wtop == top) {
      break;
    }
    if (isZenkaku(ptr[i])) {
      i++;  
    }
    wtop++;
  }
  if (wtop == top) {
    //実際の取り出し位置
    top = i+1;
  } else {
    err = ERR_VALUE;
    return;
  }
  
  // 全角を考慮した取り出し文字列の出力
  int16_t cnt=0;
  for (int16_t i = top-1 ; i < len; i++) {
    if (cnt == n) {
      break;
    }  
    c_putch(ptr[i], devno);
    if (isZenkaku(ptr[i])) {
      i++; c_putch(ptr[i], devno);
    }
    cnt++;
  }
  return;
}

// メモリ参照
// PEEK(adr)
int16_t ipeek() {
  int16_t value =0, vadr;
  uint8_t* radr;
  
  vadr = getparamValue();
  if (err) return 0;
  radr = v2realAddr(vadr);
  if (radr)
    value = *radr;
  else 
    err = ERR_VALUE;
  return value;
}

// POKEコマンド POKE ADR,データ[,データ,..データ]
void ipoke() {
  uint8_t* adr;
  int16_t value;
  int16_t vadr;
  
  // アドレスの指定
  if ( getParamRange(&vadr, 0, 32767, false) ) return;

  // 例: 1,2,3,4,5 の連続設定処理
  do {
    adr = v2realAddr(vadr);
    if (!adr) {
      err = ERR_VALUE;
      break;
    }
    cip++;          // 中間コードポインタを次へ進める
    if (getParamInt16(&value,false)) return; 
    *((uint8_t*)adr) = (uint8_t)value;
    vadr++;
  } while(*cip == I_COMMA);
}

// 美咲フォントの利用
#if USE_MISAKIFONT != 0
// フォントデータ取得
// GETFONT(仮想アドレス,SJISコード)
// 該当フォントあり 0、該当フォントなし 1
int16_t igetfont() {
  int16_t  top;  // 仮想アドレス
  uint8_t* ptr;  // フォント格納実アドレス
  uint16_t code; // 文字コード

  // 引数の取得
  if (checkOpen() ||
      getParamRange(&top, 0, 32767, true) ||
      getParam(code, false) ||
      checkClose()
  ) return 0;
  ptr  = v2realAddr(top);

  // フォントデータの取得
  return getFontDataBySJIS(ptr,  HantoZen(code));
}
#endif

/**
 * @brief '(' のチェック
 * @return '(' が正しいかどうかの判定
 * @note export gpio IR sub basic  cipはextern宣言されている
 * @todo checkparenthesis.cpp を作成して移動する cipに配慮は不要
 */
uint8_t checkOpen() {
  if (*cip != I_OPEN) err = ERR_PAREN;
  else cip++;
  return err;
}

/**
 * @brief ')' のチェック
 * @return ')' が正しいかどうかの判定
 * @note export gpio.cpp IR.cpp sub.cpp basic.cpp
 * @todo checkparenthesis.cpp を作成して移動する
 */
uint8_t checkClose() {
  if (*cip != I_CLOSE) err = ERR_PAREN;

  else cip++;
  return err;
}

/**
 * @brief 関数の引数の仮想アドレスを実アドレスに変換
 * @return NULL以外 実アドレス、NULL 範囲外を返す
 * @note export gpio.cpp sub.cpp  refer V_VAR_TOP,  in table.h 
 * @todo realadress.cpp に移動する V_VAR_TOP等は、table.h にある
 */
uint8_t* v2realAddr(uint16_t vadr) {
  uint8_t* radr = NULL; 
  if ((vadr >= V_VAR_TOP) && (vadr < V_ARRAY_TOP)) {               // 変数領域
    radr = vadr-V_VAR_TOP+(uint8_t*)var;
  } else if ((vadr >= V_ARRAY_TOP) && (vadr < V_PRG_TOP)) {        // 配列領域
    radr = vadr - V_ARRAY_TOP+(uint8_t*)arr;
  } else if ((vadr >= V_PRG_TOP) && (vadr < V_MEM_TOP)) {          // プログラム領域
    radr = vadr - V_PRG_TOP + (uint8_t*)listbuf;
  } else if ((vadr >= V_MEM_TOP) && (vadr < V_MEM_TOP+LINELEN)) {  // ユーザーワーク領域
    radr = vadr - V_MEM_TOP + (uint8_t*)lbuf;    
  } else if ((vadr >= V_MEM2_TOP) && (vadr < V_MEM2_TOP+SIZE_IBUF)) { // ユーザーワーク領域2
    radr = vadr - V_MEM2_TOP + (uint8_t*)ibuf;    
  }
  return radr;
}

// プログラム初期化
void initProgram() {
  // 変数と配列の初期化
  memset(var,0,52);
  memset(arr,0,SIZE_ARRY*2);
  gstki = 0;         // GOSUBスタックインデクスを0に初期化
  lstki = 0;         // FORスタックインデクスを0に初期化
  clp = listbuf;     // 行ポインタをプログラム保存領域の先頭に設定
  cip = clp+3;       // 中間コードポインタを先頭に設定
  val_if = 1;        // if文判定結果の初期化
}
